"""
Default instances for dynamic runs.

Typically, developers will only run one digital worker on a Python
runtime, and it is tricky to initialize an object and pass that object
instance down into other files in a project for a custom decorator.
This package provides a set of default instances of objects and contexts
for dynamic runs.

This file initializes a shared `ReportBuilder` and `Recorder` for
individual steps and sub steps to share, then creates:

    - a default ``MainContext`` named ``supervise``,
    - a decorator to annotate steps called ``step``,
    - a ``StepContext`` named ``step_context``,
    - a function to set step statuses called ``set_step_status``,
    - and a function to set record statuses called ``set_record_status``.

See the quickstart for, well, quickstarts on each of these methods.
"""

import logging
import pathlib
import warnings
from typing import Optional, Union
from urllib.parse import urlparse

import boto3
from aws_requests_auth.aws_auth import AWSRequestsAuth

from thoughtful.supervisor.environment_variables import EnvironmentVariables
from thoughtful.supervisor.main_context import MainContext
from thoughtful.supervisor.manifest import Manifest
from thoughtful.supervisor.recorder import Recorder
from thoughtful.supervisor.reporting.report_builder import ReportBuilder
from thoughtful.supervisor.step_context import StepContext
from thoughtful.supervisor.step_decorator_factory import create_step_decorator
from thoughtful.supervisor.streaming.callback import StreamingCallback

logger = logging.getLogger(__name__)
env_vars = EnvironmentVariables()

# If callback url is in environment, we will stream to that url
streaming_callback: Optional[StreamingCallback] = None
if env_vars.callback_url and env_vars.run_id:
    credentials = boto3.Session().get_credentials()
    host = urlparse(env_vars.callback_url).hostname
    aws_auth = AWSRequestsAuth(
        aws_access_key=credentials.access_key,
        aws_secret_access_key=credentials.secret_key,
        aws_token=credentials.token,
        aws_host=host,
        aws_region="us-east-1",
        aws_service="execute-api",
    )
    streaming_callback = StreamingCallback.from_env_defaults(
        auth=aws_auth, run_id=env_vars.run_id, callback_url=env_vars.callback_url
    )
else:
    logger.warning("Missing ENV variable for callback url or job id")

#: A shared ``ReportBuilder`` for steps and step contexts to add step reports
#: to.
report_builder = ReportBuilder()

#: A shared ``Recorder`` for steps and step contexts to record messages to.
recorder = Recorder()

#: Use this decorator on your own functions to mark what workflow step each
#: Python function is matched to.
step = create_step_decorator(report_builder, recorder, streaming_callback)

#: Use this function to mark a step as failed.
#: This is just a shortcut for ``set_step_status(step_id, "failed")``.
#: Exposes ``report_builder.fail_step``.
fail_step = report_builder.fail_step

#: Expose the report builder's ability to override a step's status as a
#: top-level call. Exposes ``report_builder.set_step_status``.
set_step_status = report_builder.set_step_status

#: Expose the report builder's ability to override a step's record's status as a
#: top-level call. Exposes ``report_builder.set_record_status``.
set_record_status = report_builder.set_record_status

#: Expose the report builder's ability to override a run's status as a
#: top-level call. Exposes ``report_builder.set_run_status``.
set_run_status = report_builder.set_run_status


# noinspection PyPep8Naming
class step_scope(StepContext):
    """
    It's a context manager that provides a scope for a step using the
    aforementioned default instances of ``report_builder`` and ``recorder``.
    """

    def __init__(self, *step_id, record_id: Optional[str] = None):
        """
        A default `StepContext` that uses the root level `report_builder` and
        `recorder`.

        Args:
            *step_id: The list of integers that represent the step ID.
            record_id: An optional ID of the record being actively processed
        """
        super().__init__(
            report_builder,
            recorder,
            *step_id,
            streaming_callback=streaming_callback,
            record_id=record_id,
        )


# noinspection PyPep8Naming
class substep(step_scope):
    """
    The deprecated version of ``step_scope``.
    """

    def __init__(self, *step_id):
        warnings.simplefilter("once", DeprecationWarning)
        warnings.warn(
            f"`{self.__class__.__name__}` has been renamed to `step_scope`. "
            f"Please use that class name instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        warnings.simplefilter("default", DeprecationWarning)
        super().__init__(*step_id)


# noinspection PyPep8Naming
class supervise(MainContext):
    """
    It's a context manager that provides a scope for the main context using the
    aforementioned default instances of ``report_builder`` and ``recorder``.
    """

    def __init__(
        self,
        manifest: Union[Manifest, str, pathlib.Path] = "manifest.yaml",
        output_dir: Union[str, pathlib.Path] = "output/",
        *args,
        **kwargs,
    ):
        """
        A default `MainContext` that uses the root level `report_builder` and
        `recorder`.

        Args:
            *args: Extra arguments to the `MainContext` constructor.
            **kwargs: Extra keyword arguments to the `MainContext` constructor.
            manifest (str): The digital worker's manifest definition
            output_dir (str): Where the work report and drift report will
                be written to
        """
        super().__init__(
            report_builder=report_builder,
            recorder=recorder,
            manifest=manifest,
            output_dir=output_dir,
            upload_uri=env_vars.s3_bucket_uri,
            streaming_callback=streaming_callback,
            *args,
            **kwargs,
        )



"""
The step context is a context manager that is used to supervise a single step
of a digital worker. It is used in lieu of the ``@step`` decorator for
situations when you do not want to write a function for a step. For example,
it is very useful when iterating over a list of items:

.. code-block:: python

    from thoughtful.supervisor import supervise, step_scope

    def times_two(integers: list) -> None:
        return integers * 2

    def times_three(integer: int) -> None:
        return integer * 3

    def main():
        num_list = [1, 2, 3, 4, 5]

        with step_scope("1.1"):
            num_list = times_two(num_list)

        for num in num_list:
            with step_scope("1.2"):
                num = times_three(num)

    if __name__ == '__main__':
        with supervise():
            main()

Aside: There are a few scenarios where this is actually more convenient than
using the decorator on a lower level. Thanks to the fact that we can yield
the ``self`` object from a context manager, we can use this to update
attributes such as the record status, the step status, and is even a bit
cleaner when setting the record ID.
"""

from __future__ import annotations

import copy
import warnings
from types import TracebackType
from typing import List, Optional, Type, Union

from thoughtful.supervisor.recorder import Recorder
from thoughtful.supervisor.reporting.record import Record
from thoughtful.supervisor.reporting.report_builder import ReportBuilder
from thoughtful.supervisor.reporting.report_builder import StepReportBuilder
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.timer import Timer
from thoughtful.supervisor.streaming.callback import StreamingCallback


class StepContext:
    """
    A context manager for a step that is running inside another step.
    This is an alternative to ``@step`` decorator when you don't want
    to write an entire function for a step.
    """

    def __init__(
        self,
        builder: ReportBuilder,
        recorder: Recorder,
        *step_id,
        record_id: Optional[str] = None,
        streaming_callback: Optional[StreamingCallback] = None,
    ):
        """
        Args:
            builder: Where the step report will be written.
            recorder: Where messages and data logs will be written.
            *step_id: The step id of this step, ie `"1.1"`
            record_id: An optional ID of the record being actively processed
            streaming_callback (StreamingCallback, optional): If set, streams
                the status of work report steps to that callback handler.
                Defaults to `None`.
        """
        if len(step_id) > 1:
            warnings.warn(
                "Passing multiple step ids to StepContext is deprecated. "
                "Instead, pass a single step id string with dots, ie '1.1'",
                DeprecationWarning,
            )

        self.uuid = ".".join([str(n) for n in step_id])
        self.report_builder = builder
        self.recorder = recorder
        self.timer = Timer()
        self._status_override: Optional[Status] = None
        self.record_id: Optional[str] = record_id
        self.records: List[Record] = []
        self.streaming_callback = streaming_callback
        self.step_report_builder: StepReportBuilder

    def __enter__(self):
        """
        Logic for when this context is first started.

        Returns:
            MainContext: This instance.
        """
        start_time = self.timer.start()
        record = Record(self.record_id, Status.RUNNING) if self.record_id else None

        if self.record_id:
            warnings.simplefilter("once", DeprecationWarning)
            warnings.warn(
                "Using `record_id` in a step context is deprecated and will be "
                "removed in a later release. Use `set_record_status` instead.",
                DeprecationWarning,
            )
            warnings.simplefilter("default", DeprecationWarning)

        self.step_report_builder = StepReportBuilder(
            step_id=self.uuid,
            start_time=start_time,
            status=Status.RUNNING,
            message_log=self.recorder.messages,
            data_log=self.recorder.data,
            record=record,
        )
        if self.streaming_callback:
            self.streaming_callback.post_step_update(
                self.step_report_builder.to_report()
            )
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> bool:
        """
        Runs when the context is about to close, whether caused
        by a raised Exception or now.

        Returns:
            bool: True if the parent caller should ignore the
                Exception raised before entering this function
                (if any), False otherwise.
        """
        timed_info = self.timer.end()
        if self._status_override:
            step_status = self._status_override
        else:
            step_status = Status.FAILED if exc_type else Status.SUCCEEDED

        # Update step report with finished step details
        self.step_report_builder.end_time = timed_info.end
        self.step_report_builder.status = step_status

        # Create a step report builder for every record, or create one step
        # report if no records
        final_report_builders = []
        for record in self.records:
            new_builder = copy.deepcopy(self.step_report_builder)
            new_builder.record = record
            final_report_builders.append(new_builder)

        # If there are no records, then create a "basic" step report using
        # just this instance's builder
        if not final_report_builders:
            single_step_report_builder = self.step_report_builder

            # If you're using the old way, and passing in a record ID at the step
            # context entrance, *and* you didn't override the status above, then
            # find the status based on if this context had an exception
            if self.record_id:
                record_ids_already_reported = [
                    bb.record.record_id for bb in final_report_builders
                ]
                if self.record_id not in record_ids_already_reported:
                    new_status = Status.FAILED if exc_type else Status.SUCCEEDED
                    new_record = Record(record_id=self.record_id, status=new_status)
                    single_step_report_builder.record = new_record

            final_report_builders.append(single_step_report_builder)

        # Append all step reports to the parent workflow
        self.report_builder.workflow.extend(final_report_builders)

        # Stream reports if requested
        if self.streaming_callback:
            for builder in final_report_builders:
                self.streaming_callback.post_step_update(builder.to_report())

        # Return False so that any exceptions inside this context
        # are still raised after this function ends
        return False

    def error(self) -> None:
        """
        Sets the status of this step to `Status.FAILED` in its `StepReport`.

        .. code-block:: python

            with step_scope("1.1") as s:
                ...  # do some stuff
                s.error()

        .. code-block:: json

            {
                "workflow": [
                    {
                        "step_id": "1.1",
                        "step_status": "failed"
                    }
                ]
            }
        """
        self.set_status(Status.FAILED)

    def set_status(self, status: Union[str, Status]) -> None:
        """
        Override the step context's status to be in the status of ``status``

        Args:
            status (str, Status): The status to set the step to

        .. code-block:: python

            with step_scope("1.1") as s:
                ...  # do some stuff
                s.set_status("warning")

        .. code-block:: json

            {
                "workflow": [
                    {
                        "step_id": "1.1",
                        "step_status": "warning"
                    }
                ]
            }
        """
        # Convert the status to the correct type if necessary
        safe_status = Status(status)
        self._status_override = safe_status

    def set_record_status(
        self, status: Union[str, Status], record_id: str = ""
    ) -> None:
        """
        Override a step context's record to be in the status of ``status``

        Args:
            status (str, Status): The status to set the record to
            record_id (str): The ID of the record

        .. code-block:: python

            with step_scope("1.1") as s:
                ...  # do some stuff
                s.set_record_status("warning", "record01")

        .. code-block:: json

            {
                "workflow": [
                    {
                        "step_id": "1.1",
                        "step_status": "succeeded",
                        "record": {
                            "id": "kaleb_cool_guy",
                            "status": "warning"
                        }
                    }
                ]
            }

        """
        # Convert the status to the correct type if necessary
        safe_status = Status(status)
        record_id_to_update = record_id or self.record_id

        if not record_id:
            warnings.simplefilter("once", DeprecationWarning)
            warnings.warn(
                "Setting a record status *without* the `record_id` param is "
                "deprecated and will be removed in a future release",
                DeprecationWarning,
            )
            warnings.simplefilter("default", DeprecationWarning)

        if not record_id_to_update:
            warnings.warn(
                "Setting a record status for a step without "
                "a record ID will have no effect",
                UserWarning,
            )

        new_record = Record(record_id=record_id_to_update, status=safe_status)
        self.records.append(new_record)


if __name__ == "__main__":
    report_builder = ReportBuilder()
    r = Recorder()

    substep = StepContext

    with substep(report_builder, r, 1) as s:
        print("hello world")

        with substep(report_builder, r, 1, 1) as s2:
            print("inner step")

    print(report_builder.workflow)



from __future__ import annotations

import datetime
import math
from dataclasses import dataclass
from typing import Any, Dict, Optional

import isodate


@dataclass
class TimedReport:
    """
    Jsonable report with start time, optional end time, and optional duration
    """

    start_time: datetime.datetime
    """
    datetime.datetime: The start time of the step.
    """

    end_time: Optional[datetime.datetime]
    """
    datetime.datetime: The end time of the step.
    """

    @property
    def duration(self) -> Optional[datetime.timedelta]:
        """
        datetime.timedelta: The duration of the step.
        """
        return (self.end_time - self.start_time) if self.end_time else None

    @property
    def duration_isoformat(self) -> Optional[str]:
        # Adding this because self.duration == 0 would evaluate false
        if self.duration is None:
            return None
        return isodate.duration_isoformat(self.duration)

    @property
    def duration_in_milliseconds(self) -> Optional[int]:
        # Adding this because self.duration == 0 would evaluate false
        if self.duration is None:
            return None
        return math.ceil(self.duration.total_seconds() * 1000)

    def __json__(self) -> Dict[str, Any]:
        return {
            "start_time": self.start_time.isoformat(),
            "end_time": self.end_time.isoformat() if self.end_time else None,
            "duration": self.duration_isoformat,
            "duration_in_ms": self.duration_in_milliseconds,
        }



"""
This module contains the ``Record`` class, a non-user facing class that is used
to store information about a step's record. This class is represented in the
run report as a json object.

.. code-block:: json

    {
        "workflow": [
            {
                "step_id": "1.1",
                "step_status": "succeeded",
                "record": {
                    "id": "1",
                    "status": "succeeded"
                }
            }
        ]
    }

"""

from dataclasses import dataclass

from thoughtful.supervisor.reporting.status import Status


@dataclass
class Record:
    """
    A record can refer to any object that is being processed by a step. It
    is here represented by its ID and its status.
    """

    record_id: str
    """
    str: The status of the record.
    """

    status: Status
    """
    Status: The status of the record.
    """

    def __json__(self):
        return {
            "id": self.record_id,
            "status": self.status.value,
        }



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from thoughtful.supervisor.recorder import DataLog, MessageLog
from thoughtful.supervisor.reporting.record import Record
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.timed_report import TimedReport


@dataclass
class StepReport(TimedReport):
    """
    A step report is a representation of a step's execution. This dataclass
    contains all the information that will be stored in the run report. There
    are an arbitrary number of steps reports that will be contained in the
    workflow of any given run report.
    """

    step_id: str
    """
    str: The ID of the step.
    """

    status: Status
    """
    Status: The status of the step.
    """

    message_log: MessageLog = field(default_factory=list)
    """
    MessageLog: Log of messages from the recorder object.
    """

    data_log: DataLog = field(default_factory=list)
    """
    DataLog: Log of changes to data from the recorder object.
    """

    record: Optional[Record] = None
    """
    Record, optional: The record of the step, if any.
    """

    def __json__(self) -> Dict[str, Any]:
        return {
            **super().__json__(),
            "step_id": self.step_id,
            "step_status": self.status.value,
            "data_log": self.data_log,
            "message_log": self.message_log,
            "record": self.record.__json__() if self.record else None,
        }



"""
This module contains an enum to define the actions available between
StreamingCallback and Fabric
"""


from enum import Enum


class Action(str, Enum):
    """
    Action - request payload attribute defining the nature of the request to fabric
    """

    STEP_REPORT = "step_report"
    """
    Report for new steps and updates on step status.
    """

    BOT_MANIFEST = "bot_manifest"
    """
    Provides the bots manifest so empower can contextualize step reports.
    """

    ARTIFACTS_UPLOADED = "artifacts_uploaded"
    """
    Describes the location of where a bot's run artifacts are uploaded.
    """



"""
This file is a deprecated way of getting supervisor objects.

Previously, there were two patterns of using supervisor: one was dynamic with
`supervise()`, and one was non-dynamic with `Supervisor()` and `DigitalWorker()`.

Now, there is only one pattern--dynamic--and so all of that code under the old
`dynamic/` folder is moved to the top level, because it's all that's left.

This file adds backwards compatibility for older code that uses

```
from thoughtful.supervisor.dynamic import ...
```
"""

# These are the imports to make "from thoughtful.supervisor.dynamic import ..." backwards
# compatible

import warnings

from thoughtful.supervisor.default_instances import *  # noqa

warnings.simplefilter("once", DeprecationWarning)
warnings.warn(
    "Using `from thoughtful.supervisor.dynamic import ...` is deprecated and "
    "will be removed in the future. You should use `from thoughtful.supervisor "
    "import ...` instead.",
    DeprecationWarning,
)



"""
This module contains the code for the manifest and it's dependencies. At the
base of the yaml model we have manifest metadata including UID, name,
source, author, yaml version, and default record column titles. The main
event of the manifest is the `workflow` attribute which contains a tree
of the intended workflow of the digital worker.

The `workflow` object is a list of ``Step`` yaml models. Each contains step
metadata including step ID, description, name, and overriding record column
titles. Each step can likewise contain a list of ``Step`` yaml models,
allowing for a tree of sub-steps to be defined.

For example:

.. code-block:: yaml

    uid: appnGPBr50yrJImj1
    name: CellStaff1 - Job Board Posting
    description: ""
    author: Kaleb Smith
    source: https://www.somesite.com/
    version: 0.1.0
    columns:
      succeeded: Finished
      failed: Error

    workflow:
      - step_id: 1
        title: Prepare for Process
        description: do this
        steps:
          - step_id: 1.1
            title: Get Credentials from Bitwarden
            description: do that
          - step_id: 1.2
            title: Get mapping files
            description: |-
              Get the mapping files from the server
            columns:
              succeeded: Done
              warning: Warning!
            steps:
              - step_id: 1.2.1
                title: Get mapping file from S3
      - step_id: 2
        title: Populate job postings
        description: do this
        steps:
          - step_id: 2.1
            title: Aya Connect
            description: Just your average step

The columns attributes are optional, as are each of their three values. If you
choose to override the record status column titles at the base, it will be the
default throughout the entire workflow. If you choose to override the record
status column titles at the step level, it will be the default for that step
only.
"""

from __future__ import annotations

import json
import pathlib
from collections import defaultdict
from typing import Any, List, Optional, TypeVar, Union

import pydantic
import yaml
from pydantic import validator
from pydantic_yaml import YamlModel

StepId = TypeVar("StepId", bound=str)


class RecordStatusColumns(YamlModel):
    """
    Columns is a list of column names that are used to display the
    output of a step in a table. When left blank, the default values
    are used. This is an attribute of the ``Step`` yaml model, as well as the
    base ``Manifest`` yaml model. When defined at the base, it will be the
    default for all steps. When defined at the step level, it will be the
    default for that step only.
    """

    succeeded: Optional[str]
    """
    str, optional: The name of the column that contains the succeeded records.
    """

    failed: Optional[str]
    """
    str, optional: The name of the column that contains the failed records.
    """

    warning: Optional[str]
    """
    str, optional: The name of the column that contains the warning records.
    """

    def __json__(self):
        return {
            "succeeded": self.succeeded,
            "failed": self.failed,
            "warning": self.warning,
        }


class Step(YamlModel):
    """
    A step to execute in a manifest workflow. This is an attribute of the
    ``Manifest`` yaml model. Each step can contain a list of ``Step`` yaml
    models, allowing for a tree of sub-steps to be defined.
    """

    step_id: str
    """
    str: The ID of the step.
    """

    title: str
    """
    str: The name of the step.
    """

    description: Optional[str]
    """
    str: A description of the step.
    """

    steps: Optional[List[Step]] = None
    """
    List[Step], optional: A list of children steps. Empty list if no children.
    """

    columns: Optional[RecordStatusColumns]
    """
    RecordStatusColumns, optional: The columns to display in the UI for record
    statuses. If none, it will use manifest defaults.
    """

    # PyCharm can't recognize that `@validator` creates a class method
    # noinspection PyMethodParameters
    @validator("steps", pre=True, always=True)
    def set_steps(cls, steps):
        """
        Validator: Set the steps attribute to an empty list if it is None.
        """
        return steps or []

    def __json__(self):
        return {
            "step_id": self.step_id,
            "title": self.title,
            "description": self.description,
            "steps": [step.__json__() for step in self.steps],
            "columns": self.columns.__json__() if self.columns else self.columns,
        }


class Manifest(YamlModel):
    """
    A digital worker manifest, typically read from a `manifest.yaml` file.
    """

    uid: str
    """
    str: The unique ID of the manifest.
    """

    name: str
    """
    str: The name of the process.
    """

    description: str
    """
    str: Manifest/process description.
    """

    author: Optional[str]
    """
    str, optional: Who wrote the manifest.
    """

    source: str
    """
    str: Where the manifest came from.
    """

    columns: Optional[RecordStatusColumns]
    """
    RecordStatusColumns, optional: The columns to display in the UI for record
    statuses. If none, it will use defaults.
    """

    workflow: List[Step]
    """
    List[Step]: The list of steps to execute.
    """

    @classmethod
    def from_file(cls, filename: Union[str, pathlib.Path]) -> Manifest:
        """
        Creates an instance of this class from a YAML file.

        Args:
            filename: The name of the manifest .yaml file.

        Returns:
            Manifest: The new Manifest instance.
        """
        # Load the manifest using the base class importer
        try:
            manifest_in_flight = cls.parse_file(str(filename))
        except pydantic.error_wrappers.ValidationError as ex:
            errors = ex.errors()
            error_message = (
                f"Invalid manifest file with {len(errors)} "
                f"validation error(s) at the root "
                f"level\n{_pretty_errors_string(errors)}"
            )
            raise ValueError(error_message) from ex

        # The base importer loads the steps as dicts, so convert them into
        # proper steps
        step_dicts = manifest_in_flight.workflow
        final_steps: List[Step] = cls._load_manifest_file_steps(step_dicts)
        manifest_in_flight.workflow = final_steps

        return manifest_in_flight

    @classmethod
    def yaml_to_json(
        cls,
        manifest_path: Union[str, pathlib.Path],
        json_path: Union[str, pathlib.Path],
    ) -> None:
        """
        Converts a YAML manifest into a JSON manifest by converting the raw YAML
        contents into a JSON dictionary. This function is "dumb" and just writes
        the direct-to-JSON version of the YAML file.

        Args:
            manifest_path: The manifest YAML to consume.
            json_path: The output file to write the JSON payload to.
        """
        with manifest_path.open("r") as f:
            manifest_yaml = yaml.safe_load(f)

        with pathlib.Path(json_path).open("w") as json_path:
            json.dump(manifest_yaml, json_path)

    def __json__(self) -> dict[str, Any]:
        """
        Returns the values of this instance as a dictionary that are needed
        for a report.
        Returns:
            dict: The JSON-able dictionary (of only primitives) representation
                of this instance.
        """
        # Only return the keys we want in the report
        keys = ["uid", "name", "description", "author", "source", "columns"]
        manifest_json = {k: v for k, v in self.__dict__.items() if k in keys}
        manifest_json["workflow"] = [step.__json__() for step in self.workflow]
        return manifest_json

    def write_to_json_file(self, json_path):
        _json = self.__json__()
        with open(json_path, "w+") as json_file:
            json.dump(_json, json_file)

    # Private implementation
    @classmethod
    def _load_manifest_file_steps(
        cls, step_dicts: List[Union[dict, Step]]
    ) -> List[Step]:
        """
        Creates a list of Step instances based on a list of steps in dictionary
        format (usually loaded from a raw YAML file).

        Args:
            step_dicts: The steps as dictionaries to parse.

        Returns:
            List[Step]: The new Step instances.

        """
        successful_steps = []
        # Keep track of errors for specific step IDs, and also errors for any
        # steps that don't have a step ID
        errors_for_step_ids: dict[str, List[Any]] = defaultdict(list)
        errors_for_unknown_steps: List[Any] = []

        # Try parsing each step
        for step_dict in step_dicts:
            try:
                new_step = Step.parse_obj(step_dict)
                successful_steps.append(new_step)
            except pydantic.ValidationError as ex:
                errors = ex.errors()
                if "step_id" in step_dict:
                    _id = step_dict["step_id"]
                    errors_for_step_ids[_id].extend(errors)
                else:
                    errors_for_unknown_steps.append(errors)

        # Back out if there's no errors
        if not (errors_for_step_ids or errors_for_unknown_steps):
            return successful_steps

        # Otherwise, build the pretty print error string
        message = "Manifest step(s) are invalid"
        for step_id, errors in errors_for_step_ids.items():
            message += f"\nstep id: {step_id}\n"
            message += _pretty_errors_string(errors, indent=4)

        for index, step_errors in enumerate(errors_for_unknown_steps):
            message += f"\nunknown step {index + 1} (no step ID found)\n"
            message += _pretty_errors_string(step_errors, indent=4)

        raise ValueError(message)


def _pretty_errors_string(errors: list, indent: int = 0) -> str:
    """
    A custom string formatter for pydantic errors.

    See Also:
        The original implementation of this (with a slightly different format)
        is here - https://github.com/samuelcolvin/pydantic/blob/8846ec4685e749b93907081450f592060eeb99b1/pydantic/error_wrappers.py#L82-L83
    """

    def pretty_error(error: dict) -> str:
        loc_str = " -> ".join(str(ll) for ll in error["loc"])
        spacer = " " * indent
        return f'{spacer}{loc_str}\n    {spacer}{error["msg"]}'

    return "\n".join(pretty_error(err) for err in errors)
